/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*
	Author: Alain Gibaud 2012
	alain.gibaud@free.fr
*/
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <cstdio>
#include <time.h>
#include <ctype.h>
#include <cstdio>
#include <cstdlib>

#include "processor.h"

using namespace std ;


int execute_command(const string& comm, vector<string>&  r) ;
bool locate_inc() ;
string get_path(const string & path_and_file) ;
string get_filename(const string & path_and_file) ;
string get_nakedfilename(const string &  file) ;
string extract(const string &source , int pmode) ;
string extract_all(const string &dir, int pmode) ;

bool decode_arg(int argc, char *argv[]) ;

// --------------------------------
int mode ; 		// 0 = unknown, 1 = process one file, 2 = locate, 3 = process a directory
int printmode ; 	// 1 = only structs, 2=only defines, 3=both
string arg ; 		// file or directory name
// --------------------------------

bool decode_arg(int argc, char *argv[])
{
    int i ;
    mode = 0 ;
    printmode = 0 ;
    arg.clear() ;

    for( i = 1 ; i < argc ; ++i)
    {
        if(argv[i] == string("-locate"))
        {
            if( mode != 0) return false ;
            mode = 2 ;
        }
        else if( argv[i] == string("-all"))
        {
            if( mode != 0 || i+1 >= argc) return false ;
            arg = argv[++i] ;
            mode = 3 ;
        }
	else if(  argv[i] == string("-define"))
	{
		printmode |= 2 ;
	}
        else if( argv[i] == string("-struct"))
	{
		printmode |= 1 ;
	}
        else
        {
            if( mode != 0) return false ;
            arg = argv[i] ;
            mode = 1 ;
        }
    }
    if(printmode == 0) printmode = 3 ;
    return true ;
}

void usage()
{
    cout << "usage:    inc2h  [options]  <inc_file>" << endl ;
    cout << "or        inc2h  [options]  -all <directory>" << endl ;
    cout << "or        inc2h -locate" << endl ;
    cout << "options   :" << endl ;
    cout << "-struct   : generates Microchip-compatible structs" << endl ;
    cout << "-define   : generates #defines compatible with cpik explicit bit-fields" << endl ;
    cout << "( by defaults, both are generated )" << endl ;
    exit(1) ;
}

int main(int argc, char *argv[])
{
    string status ;

    if( !decode_arg(argc, argv) ) usage() ;

    switch(mode)
    {
    case 0 :
        usage() ;
        break ;

    case 1:
        status = extract(arg,printmode) ;
        if( status != "")
        {
            cerr << status  << endl ;
            exit(2) ;
        }
        break ;

    case 2:
        status = locate_inc() ;
        if( status != "")
        {
            cerr << status  << endl ;
            exit(2) ;
        }
        break ;
    case 3:
        status = extract_all(arg,printmode) ;
        if( status != "")
        {
            cerr << status  << endl ;
            exit(2) ;
        }
        break ;
    }

    return 0 ;
}

// -------------------------------------------------------

int execute_command(const string& comm, vector<string>&  r)
{
    int status ;
    char t[512] ;
    FILE *f ;

    r.clear() ;
    f = popen((comm   + " 2>&1"  ).c_str(),"r") ;
    if( f == 0) return -1;

    while(fgets(t,511,f) == t)
    {
        unsigned int i = strlen(t) - 1 ;
        if(i >= 0 && t[i] == '\n') t[i] = 0 ;
        r.push_back( string(t) );
    }
    status = pclose(f) ;
    return status ;
}

// -------------------------------------------------------


bool locate_inc()
{
    vector<string> r ;
    string dir ;

    int s = execute_command("locate p18f1220.inc", r) ;

    if( s != 0 || r.size() == 0) return false ;

    for( unsigned int i = 0 ; i < r.size() ; ++i)
    {
        if( r[i].size() != 0 && r[i][0] == '/')
        {
            dir = get_path(r[i]) ;
            cout << dir << endl ;
        }
    }

    return true ;
}


// -------------------------------------------------------


string get_path(const string & path_and_file)
{

    int i = path_and_file.rfind("/") ;
    return path_and_file.substr(0,i+1) ; ;
}

// -------------------------------------------------------


string get_filename(const string & path_and_file)
{
    int i = path_and_file.rfind("/") ;
    return path_and_file.substr(i+1) ; ;
}


// -------------------------------------------------------

string get_nakedfilename(const string &  file)
{
    int i =  file.rfind(".") ;
    return file.substr(0,i) ; ;
}

// -------------------------------------------------------
string extract(const string &source,int pmode)
{

    string error ;
    ifstream input(source.c_str()) ;



    if( ! input.is_open())
    {
        return  "Can't open " + source ;
    }


    string naked = get_nakedfilename( get_filename(source) );
    naked = bfblock::upper(naked) ;
    string dest = naked +".h" ;
    ofstream output(dest.c_str()) ;

    if( ! output.is_open())
    {
        return  "Can't open " + dest;
    }

    processor proc (naked ) ;


    if( !proc.extract_processor( input))
    {
        return "Can't find processor name"   ;
    }

    if( !proc.extract_registers( input ))
    {
        return "Can't find register definitions"   ;
    }

    if( !proc.extract_bits( input))
    {
        return "Can't find bit definitions"  ;
    }

    proc.make_aliases() ;
    proc.setmode(pmode) ;
    proc.make_H(output) ;

    return error ;
}


// -------------------------------------------------------

string extract_all(const string &dir,int pmode)
{
    string status ;
    vector<string> r ;

    int s = execute_command("ls -1 "+dir+"p18*.inc "+dir+"P18*.INC", r) ;

    if( s == 0 && r.size() != 0)
    {
        for( unsigned int i = 0 ; status == "" && i < r.size() ; ++i)
        {
            cout << "Extracting : " << r[i] << endl << flush;
            status = extract(r[i], pmode) ;
        }
    }

    return status ;
}
